from typing import Type

import lucid.nn as nn
import math

import lucid
from lucid import register_model
from lucid._tensor import Tensor


__all__ = [
    "EfficientNet",
    "EfficientNet_V2",
    "efficientnet_b0",
    "efficientnet_b1",
    "efficientnet_b2",
    "efficientnet_b3",
    "efficientnet_b4",
    "efficientnet_b5",
    "efficientnet_b6",
    "efficientnet_b7",
    "efficientnet_v2_s",
    "efficientnet_v2_m",
    "efficientnet_v2_l",
    "efficientnet_v2_xl",
]


def _make_divisible(v: int, divisor: int, min_value: int | None = None) -> int:
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)

    if new_v < 0.9 * v:
        new_v += divisor

    return int(new_v)


class _SEBlock(nn.Module):
    def __init__(self, in_channels: int, mid_channels: int) -> None:
        super().__init__()
        self.squeeze = nn.AdaptiveAvgPool2d((1, 1))
        self.excite = nn.Sequential(
            nn.Linear(in_channels, mid_channels),
            nn.Swish(),
            nn.Linear(mid_channels, in_channels),
            nn.Sigmoid(),
        )

    def forward(self, x: Tensor) -> Tensor:
        x = self.squeeze(x).reshape(x.shape[0], -1)
        x = self.excite(x).unsqueeze(axis=(-1, -2))
        return x


class _MBConv(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int = 1,
        se_scale: int = 4,
        expansion: int = 6,
        p: float = 0.5,
    ) -> None:
        super().__init__()
        self.p = p if in_channels == out_channels else 1.0
        self.shortcut = stride == 1 and in_channels == out_channels
        self.expansion = expansion

        self.expand = nn.Sequential(
            nn.Conv2d(
                in_channels,
                in_channels * expansion,
                kernel_size=1,
                stride=stride,
                bias=False,
            ),
            nn.BatchNorm2d(in_channels * expansion, eps=1e-3),
            nn.Swish(),
        )
        self.residual = nn.Sequential(
            nn.Conv2d(
                in_channels * expansion,
                in_channels * expansion,
                kernel_size=kernel_size,
                padding="same",
                groups=in_channels * expansion,
                bias=False,
            ),
            nn.BatchNorm2d(in_channels * expansion, eps=1e-3),
            nn.Swish(),
        )

        self.se = _SEBlock(in_channels * expansion, max(1, int(in_channels / se_scale)))
        self.project = nn.Sequential(
            nn.Conv2d(in_channels * expansion, out_channels, kernel_size=1, bias=False),
            nn.BatchNorm2d(out_channels, eps=1e-3),
        )

    def forward(self, x: Tensor) -> Tensor:
        if self.training:
            if not lucid.random.bernoulli(self.p).item():
                return x

        x_shortcut = x
        x_residual = self.expand(x) if self.expansion != 1 else x
        x_residual = self.residual(x_residual)
        x_se = self.se(x_residual)

        x = x_se * x_residual
        x = self.project(x)

        if self.shortcut:
            x = x_shortcut + x
        return x


class EfficientNet(nn.Module):
    def __init__(
        self,
        num_classes: int = 1000,
        width_coef: float = 1.0,
        depth_coef: float = 1.0,
        scale: float = 1.0,
        dropout: float = 0.2,
        se_scale: int = 4,
        stochastic_depth: bool = False,
        p: float = 0.5,
    ) -> None:
        super().__init__()
        channels = [32, 16, 24, 40, 80, 112, 192, 320, 1280]
        repeats = [1, 2, 2, 3, 3, 4, 1]
        strides = [1, 2, 2, 2, 1, 2, 1]
        kernel_sizes = [3, 3, 5, 3, 5, 5, 3]
        expands = [1, 6, 6, 6, 6, 6, 6]

        depth = depth_coef
        width = width_coef

        channels = [_make_divisible(ch * width, 8) for ch in channels]
        repeats = [math.ceil(rep * depth) for rep in repeats]

        if stochastic_depth:
            self.p = p
            self.step = 0.5 / (sum(repeats - 1))
        else:
            self.p = 1.0
            self.step = 0.0

        self.upsample = nn.Upsample(
            scale_factor=scale, mode="bilinear", align_corners=False
        )

        self.stage1 = nn.Sequential(
            nn.Conv2d(3, channels[0], kernel_size=3, stride=2, padding=1, bias=False),
            nn.BatchNorm2d(channels[0], eps=1e-3),
        )

        for i in range(7):
            block = self._make_block(
                _MBConv,
                repeats[i],
                channels[i],
                channels[i + 1],
                kernel_sizes[i],
                strides[i],
                expands[i],
                se_scale,
            )
            self.add_module(f"stage{i + 2}", block)

        self.stage9 = nn.Sequential(
            nn.Conv2d(channels[7], channels[8], kernel_size=1, bias=False),
            nn.BatchNorm2d(channels[8], eps=1e-3),
            nn.Swish(),
        )

        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.dropout = nn.Dropout(p=dropout)
        self.fc = nn.Linear(channels[8], num_classes)

    def _make_block(
        self,
        block: Type[nn.Module],
        repeats: int,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int,
        expansion: int,
        se_scale: int,
    ) -> nn.Sequential:
        strides = [stride] + [1] * (repeats - 1)
        layers = []
        for stride in strides:
            layers.append(
                block(
                    in_channels,
                    out_channels,
                    kernel_size,
                    stride,
                    se_scale,
                    expansion,
                    self.p,
                )
            )
            in_channels = out_channels
            self.p -= self.step

        return nn.Sequential(*layers)

    def forward(self, x: Tensor) -> Tensor:
        x = self.upsample(x)
        for i in range(1, 10):
            x = getattr(self, f"stage{i}")(x)

        x = self.avgpool(x)
        x = x.reshape(x.shape[0], -1)
        x = self.dropout(x)
        x = self.fc(x)

        return x


class _FusedMBConv(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int = 1,
        se_scale: int = 0,
        expansion: int = 4,
        drop_prob: float = 0.2,
    ) -> None:
        super().__init__()
        self.drop_prob = drop_prob
        self.has_shortcut = stride == 1 and in_channels == out_channels

        mid_channels = in_channels * expansion
        layers = []

        if expansion != 1:
            layers.append(
                nn.Conv2d(
                    in_channels,
                    mid_channels,
                    kernel_size,
                    stride,
                    padding="same",
                    bias=False,
                )
            )
            layers.append(nn.BatchNorm2d(mid_channels, eps=1e-3))
            layers.append(nn.Swish())

            layers.append(
                nn.Conv2d(mid_channels, out_channels, kernel_size=1, bias=False)
            )
            layers.append(nn.BatchNorm2d(out_channels, eps=1e-3))

        else:
            layers.append(
                nn.Conv2d(
                    in_channels,
                    out_channels,
                    kernel_size,
                    stride,
                    padding="same",
                    bias=False,
                )
            )
            layers.append(nn.BatchNorm2d(out_channels, eps=1e-3))
            layers.append(nn.Swish())

        self.fused_conv = nn.Sequential(*layers)
        if se_scale > 0:
            self.se = _SEBlock(out_channels, max(1, in_channels // se_scale))
        else:
            self.se = nn.Identity()

    def forward(self, x: Tensor) -> Tensor:
        if self.training:
            if not lucid.random.bernoulli(self.drop_prob).item():
                return x

        out = self.fused_conv(x)
        out = self.se(out)
        if self.has_shortcut:
            out += x

        return out


class EfficientNet_V2(nn.Module):
    def __init__(
        self,
        block_cfg: list,
        num_classes: int = 1000,
        dropout: float = 0.2,
        drop_path_rate: float = 0.2,
    ) -> None:
        super().__init__()
        stem_out = block_cfg[0][1]
        self.stem = nn.Sequential(
            nn.Conv2d(3, stem_out, kernel_size=3, stride=2, padding=1, bias=False),
            nn.BatchNorm2d(stem_out, eps=1e-3),
            nn.Swish(),
        )

        block_layers = []
        total_repeats = sum(cfg[5] for cfg in block_cfg)
        dp_index = 0

        in_channels = stem_out
        for fused, cout, k, s, e, r, se_scale in block_cfg:
            for i in range(r):
                drop_prob = 1.0 - (dp_index / total_repeats) * drop_path_rate
                dp_index += 1

                if fused:
                    block = _FusedMBConv(
                        in_channels, cout, k, s if i == 0 else 1, se_scale, e, drop_prob
                    )
                else:
                    block = _MBConv(
                        in_channels, cout, k, s if i == 0 else 1, se_scale, e, drop_prob
                    )

                block_layers.append(block)
                in_channels = cout

        self.features = nn.Sequential(*block_layers)

        head_ch = in_channels * 4
        self.head = nn.Sequential(
            nn.Conv2d(in_channels, head_ch, kernel_size=1, bias=False),
            nn.BatchNorm2d(head_ch, eps=1e-3),
            nn.Swish(),
        )

        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.dropout = nn.Dropout(p=dropout)
        self.fc = nn.Linear(head_ch, num_classes)

    def forward(self, x: Tensor) -> Tensor:
        x = self.stem(x)
        x = self.features(x)
        x = self.head(x)
        x = self.avgpool(x)

        x = x.reshape(x.shape[0], -1)
        x = self.dropout(x)
        x = self.fc(x)

        return x


@register_model
def efficientnet_b0(num_classes: int = 1000, **kwargs) -> EfficientNet:
    return EfficientNet(
        num_classes,
        width_coef=1.0,
        depth_coef=1.0,
        scale=224 / 224,
        dropout=0.2,
        **kwargs,
    )


@register_model
def efficientnet_b1(num_classes: int = 1000, **kwargs) -> EfficientNet:
    return EfficientNet(
        num_classes,
        width_coef=1.0,
        depth_coef=1.1,
        scale=240 / 224,
        dropout=0.2,
        **kwargs,
    )


@register_model
def efficientnet_b2(num_classes: int = 1000, **kwargs) -> EfficientNet:
    return EfficientNet(
        num_classes,
        width_coef=1.1,
        depth_coef=1.2,
        scale=260 / 224,
        dropout=0.3,
        **kwargs,
    )


@register_model
def efficientnet_b3(num_classes: int = 1000, **kwargs) -> EfficientNet:
    return EfficientNet(
        num_classes,
        width_coef=1.2,
        depth_coef=1.4,
        scale=300 / 224,
        dropout=0.3,
        **kwargs,
    )


@register_model
def efficientnet_b4(num_classes: int = 1000, **kwargs) -> EfficientNet:
    return EfficientNet(
        num_classes,
        width_coef=1.4,
        depth_coef=1.8,
        scale=380 / 224,
        dropout=0.4,
        **kwargs,
    )


@register_model
def efficientnet_b5(num_classes: int = 1000, **kwargs) -> EfficientNet:
    return EfficientNet(
        num_classes,
        width_coef=1.6,
        depth_coef=2.2,
        scale=456 / 224,
        dropout=0.4,
        **kwargs,
    )


@register_model
def efficientnet_b6(num_classes: int = 1000, **kwargs) -> EfficientNet:
    return EfficientNet(
        num_classes,
        width_coef=1.8,
        depth_coef=2.6,
        scale=528 / 224,
        dropout=0.5,
        **kwargs,
    )


@register_model
def efficientnet_b7(num_classes: int = 1000, **kwargs) -> EfficientNet:
    return EfficientNet(
        num_classes,
        width_coef=2.0,
        depth_coef=3.1,
        scale=600 / 224,
        dropout=0.5,
        **kwargs,
    )


@register_model
def efficientnet_v2_s(num_classes: int = 1000, **kwargs) -> EfficientNet_V2:
    cfg = [
        [True, 24, 3, 2, 1, 2, 0],
        [True, 48, 3, 2, 4, 4, 0],
        [True, 64, 3, 2, 4, 4, 0],
        [False, 128, 3, 2, 4, 6, 4],
        [False, 160, 3, 1, 6, 9, 4],
        [False, 256, 3, 2, 6, 15, 4],
    ]
    return EfficientNet_V2(cfg, num_classes, dropout=0.2, drop_path_rate=0.2, **kwargs)


@register_model
def efficientnet_v2_m(num_classes: int = 1000, **kwargs) -> EfficientNet_V2:
    cfg = [
        [True, 24, 3, 2, 1, 3, 0],
        [True, 48, 3, 2, 4, 5, 0],
        [True, 80, 3, 2, 4, 5, 0],
        [False, 160, 3, 2, 4, 7, 4],
        [False, 176, 3, 1, 6, 14, 4],
        [False, 304, 3, 2, 6, 18, 4],
        [False, 512, 3, 1, 6, 5, 4],
    ]
    return EfficientNet_V2(cfg, num_classes, dropout=0.3, drop_path_rate=0.2, **kwargs)


@register_model
def efficientnet_v2_l(num_classes: int = 1000, **kwargs) -> EfficientNet_V2:
    cfg = [
        [True, 32, 3, 2, 1, 4, 0],
        [True, 64, 3, 2, 4, 7, 0],
        [True, 96, 3, 2, 4, 7, 0],
        [False, 192, 3, 2, 4, 10, 4],
        [False, 224, 3, 1, 6, 19, 4],
        [False, 384, 3, 2, 6, 25, 4],
        [False, 640, 3, 1, 6, 7, 4],
    ]
    return EfficientNet_V2(cfg, num_classes, dropout=0.4, drop_path_rate=0.3, **kwargs)


@register_model
def efficientnet_v2_xl(num_classes: int = 1000, **kwargs) -> EfficientNet_V2:
    cfg = [
        [True, 32, 3, 2, 1, 4, 0],
        [True, 64, 3, 2, 4, 8, 0],
        [True, 96, 3, 2, 4, 8, 0],
        [False, 192, 3, 2, 4, 16, 4],
        [False, 256, 3, 1, 6, 24, 4],
        [False, 512, 3, 2, 6, 32, 4],
        [False, 640, 3, 1, 6, 8, 4],
    ]
    return EfficientNet_V2(cfg, num_classes, dropout=0.5, drop_path_rate=0.4, **kwargs)



from typing import Type

import lucid
import lucid.nn as nn

from lucid import register_model
from lucid._tensor import Tensor


__all__ = [
    "CoAtNet",
    "coatnet_0",
    "coatnet_1",
    "coatnet_2",
    "coatnet_3",
    "coatnet_4",
    "coatnet_5",
    "coatnet_6",
    "coatnet_7",
]


def conv_3x3_bn(
    in_channels: int, out_channels: int, downsample: bool = False, **kwargs
) -> nn.Sequential:
    stride = 1 if not downsample else 2
    return nn.Sequential(
        nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=3,
            stride=stride,
            padding=1,
            bias=False,
        ),
        nn.BatchNorm2d(out_channels),
        nn.GELU(),
    )


class _PreNorm(nn.Module):
    def __init__(self, dim: int, fn: nn.Module, norm: Type[nn.Module]) -> Tensor:
        super().__init__()
        self.norm = norm(dim)
        self.fn = fn

    def forward(self, x: Tensor, **kwargs) -> Tensor:
        return self.fn(self.norm(x), **kwargs)


class _SEBlock(nn.Module):
    def __init__(
        self, in_channels: int, out_channels: int, reduction: float = 0.25
    ) -> None:
        super().__init__()
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Sequential(
            nn.Linear(out_channels, int(in_channels * reduction), bias=False),
            nn.GELU(),
            nn.Linear(int(in_channels * reduction), out_channels, bias=False),
            nn.Sigmoid(),
        )

    def forward(self, x: Tensor) -> Tensor:
        y = self.avgpool(x)
        y = y.reshape(y.shape[0], -1)
        y = self.fc(y).unsqueeze(axis=(-1, -2))

        return x * y


class _FeedForward(nn.Module):
    def __init__(self, dim: int, hidden_dim: int, dropout: float = 0.0) -> None:
        super().__init__()
        self.ff = nn.Sequential(
            nn.Linear(dim, hidden_dim),
            nn.GELU(),
            nn.Dropout(dropout),
            nn.Linear(hidden_dim, dim),
            nn.Dropout(dropout),
        )

    def forward(self, x: Tensor) -> Tensor:
        return self.ff(x)


class _MBConv(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        downsample: bool = False,
        expansion: int = 4,
        **kwargs
    ) -> None:
        super().__init__()
        self.downsample = downsample
        stride = 1 if not self.downsample else 2
        hidden_dim = int(in_channels * expansion)

        if self.downsample:
            self.pool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
            self.proj = nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=False)

        if expansion == 1:
            conv = nn.Sequential(
                nn.Conv2d(
                    hidden_dim,
                    hidden_dim,
                    kernel_size=3,
                    stride=stride,
                    padding=1,
                    groups=hidden_dim,
                    bias=False,
                ),
                nn.BatchNorm2d(hidden_dim),
                nn.GELU(),
                nn.Conv2d(hidden_dim, out_channels, kernel_size=1, bias=False),
                nn.BatchNorm2d(out_channels),
            )
        else:
            conv = nn.Sequential(
                nn.Conv2d(
                    in_channels, hidden_dim, kernel_size=1, stride=stride, bias=False
                ),
                nn.BatchNorm2d(hidden_dim),
                nn.GELU(),
                nn.Conv2d(
                    hidden_dim,
                    hidden_dim,
                    kernel_size=3,
                    padding=1,
                    groups=hidden_dim,
                    bias=False,
                ),
                nn.BatchNorm2d(hidden_dim),
                nn.GELU(),
                _SEBlock(in_channels, hidden_dim),
                nn.Conv2d(hidden_dim, out_channels, kernel_size=1, bias=False),
                nn.BatchNorm2d(out_channels),
            )

        self.conv = _PreNorm(in_channels, conv, nn.BatchNorm2d)

    def forward(self, x: Tensor) -> Tensor:
        if self.downsample:
            return self.proj(self.pool(x)) + self.conv(x)
        else:
            return x + self.conv(x)


class _Attention(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        img_size: tuple[int, int],
        num_heads: int = 8,
        dim_head: int = 32,
        dropout: float = 0.0,
    ) -> None:
        super().__init__()
        inner_dim = dim_head * num_heads
        project_out = not (num_heads == 1 and dim_head == in_channels)

        self.ih, self.iw = img_size
        self.num_heads = num_heads
        self.scale = dim_head**0.5

        self.relative_bias_table = nn.Parameter(
            lucid.zeros((2 * self.ih - 1) * (2 * self.iw - 1), num_heads)
        )

        coords = lucid.stack(
            lucid.meshgrid(lucid.arange(self.ih), lucid.arange(self.iw))
        )
        coords = coords.reshape(coords.shape[0], -1)
        relative_coords = coords[:, :, None] - coords[:, None, :]

        relative_coords[0] += self.ih - 1
        relative_coords[1] += self.iw - 1
        relative_coords[0] *= 2 * self.iw - 1

        relative_coords = lucid.einops.rearrange(relative_coords, "c h w -> h w c")
        relative_index = relative_coords.sum(axis=-1).flatten().unsqueeze(axis=1)

        self.relative_index: nn.Buffer
        self.register_buffer("relative_index", relative_index)

        self.attend = nn.Softmax(axis=-1)
        self.to_qkv = nn.Linear(in_channels, inner_dim * 3, bias=False)

        self.to_out = (
            nn.Sequential(nn.Linear(inner_dim, out_channels), nn.Dropout(dropout))
            if project_out
            else nn.Identity()
        )

    def forward(self, x: Tensor) -> Tensor:
        qkv = self.to_qkv(x).chunk(3, axis=-1)
        q, k, v = map(
            lambda tensor: lucid.einops.rearrange(
                tensor,
                "b n (h d) -> b h n d",
                h=self.num_heads,
                d=tensor.shape[-1] // self.num_heads,
            ),
            qkv,
        )
        dots = (q @ k.mT) * self.scale

        rel_idx = self.relative_index.repeat(self.num_heads, axis=1).to(x.device)
        col_indices = (
            (
                lucid.arange(self.relative_bias_table.shape[1])
                .unsqueeze(axis=0)
                .broadcast_to(rel_idx.shape)
            )
            .astype(lucid.Int)
            .to(x.device)
        )

        relative_bias = self.relative_bias_table[
            rel_idx.astype(lucid.Int), col_indices.astype(lucid.Int)
        ]
        relative_bias = lucid.einops.rearrange(
            relative_bias, "(h w) c -> c h w", h=self.ih * self.iw, w=self.ih * self.iw
        ).unsqueeze(axis=0)

        dots += relative_bias

        attn = self.attend(dots)
        out = attn @ v
        out = lucid.einops.rearrange(out, "b h n d -> b n (h d)")
        out = self.to_out(out)

        return out


class _Transformer(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        img_size: tuple[int, int],
        num_heads: int = 8,
        dim_head: int = 32,
        downsample: bool = False,
        dropout: float = 0.0,
    ) -> None:
        super().__init__()
        hidden_dim = int(in_channels * 4)

        self.ih, self.iw = img_size
        self.downsample = downsample

        if self.downsample:
            self.pool1 = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
            self.pool2 = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
            self.proj = nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=False)

        attn = _Attention(
            in_channels, out_channels, img_size, num_heads, dim_head, dropout
        )
        ff = _FeedForward(out_channels, hidden_dim, dropout)

        self.attn = nn.Sequential(
            nn.Rearrange("b c ih iw -> b (ih iw) c"),
            _PreNorm(in_channels, attn, nn.LayerNorm),
            nn.Rearrange("b (ih iw) c -> b c ih iw", ih=self.ih, iw=self.iw),
        )
        self.ff = nn.Sequential(
            nn.Rearrange("b c ih iw -> b (ih iw) c"),
            _PreNorm(out_channels, ff, nn.LayerNorm),
            nn.Rearrange("b (ih iw) c -> b c ih iw", ih=self.ih, iw=self.iw),
        )

    def forward(self, x: Tensor) -> Tensor:
        if self.downsample:
            x = self.proj(self.pool1(x)) + self.attn(self.pool2(x))
        else:
            x += self.attn(x)
        x += self.ff(x)
        return x


class CoAtNet(nn.Module):
    def __init__(
        self,
        img_size: tuple[int, int],
        in_channels: int,
        num_blocks: list[int],
        channels: list[int],
        num_classes: int = 1000,
        num_heads: int = 32,
        block_types: list[str] = ["C", "C", "T", "T"],
        _scaled_num_blocks: list[int] | None = None,
        _scaled_channels: list[int] | None = None,
    ) -> None:
        super().__init__()
        block = {"C": _MBConv, "T": _Transformer}
        get_block = lambda i: block[block_types[i]]

        self.ih, self.iw = img_size
        self.channels = channels
        self.num_heads = num_heads

        self._do_scale = _scaled_num_blocks is not None and _scaled_channels is not None

        self.s0 = self._make_layer(
            0, conv_3x3_bn, in_channels, channels[0], num_blocks[0]
        )
        self.s1 = self._make_layer(
            1, get_block(0), channels[0], channels[1], num_blocks[1]
        )
        self.s2 = self._make_layer(
            2, get_block(1), channels[1], channels[2], num_blocks[2]
        )
        self.s3 = (
            self._make_layer(3, get_block(2), channels[2], channels[3], num_blocks[3])
            if not self._do_scale
            else nn.Identity()
        )
        self.s4 = self._make_layer(
            4,
            get_block(3),
            channels[3] if not self._do_scale else _scaled_channels[1],
            channels[4],
            num_blocks[4],
        )

        self.pool = nn.AvgPool2d(kernel_size=self.ih // 32)
        self.fc = nn.Linear(channels[-1], num_classes, bias=False)

        if self._do_scale:
            self._scale_s3(_scaled_num_blocks, _scaled_channels)

    def _make_layer(
        self,
        i: int,
        block: Type[nn.Module],
        in_channels: int,
        out_channels: int,
        depth: int,
    ) -> nn.Sequential:
        img_size = (self.ih // (2 ** (i + 1)), self.iw // (2 ** (i + 1)))

        layers = nn.ModuleList()
        for i in range(depth):
            new_block = block(
                in_channels if i == 0 else out_channels,
                out_channels,
                img_size=img_size,
                num_heads=self.num_heads,
                downsample=True if i == 0 else False,
            )
            layers.append(new_block)

        return nn.Sequential(*layers)

    def _scale_s3(self, depths: list[int, int], channels: list[int, int]) -> None:
        self.s3_tandem = nn.Sequential(
            self._make_layer(3, _MBConv, self.channels[2], channels[0], depths[0]),
            self._make_layer(3, _Transformer, channels[0], channels[1], depths[1]),
        )

    def forward(self, x: Tensor) -> Tensor:
        x = self.s0(x)
        x = self.s1(x)
        x = self.s2(x)
        x = self.s3(x) if not self._do_scale else self.s3_tandem(x)
        x = self.s4(x)

        x = self.pool(x).reshape(-1, x.shape[1])
        x = self.fc(x)

        return x


@register_model
def coatnet_0(num_classes: int = 1000, **kwargs) -> CoAtNet:
    num_blocks = [2, 2, 3, 5, 2]
    channels = [64, 96, 192, 384, 768]
    return CoAtNet((224, 224), 3, num_blocks, channels, num_classes, **kwargs)


@register_model
def coatnet_1(num_classes: int = 1000, **kwargs) -> CoAtNet:
    num_blocks = [2, 2, 6, 14, 2]
    channels = [64, 96, 192, 384, 768]
    return CoAtNet((224, 224), 3, num_blocks, channels, num_classes, **kwargs)


@register_model
def coatnet_2(num_classes: int = 1000, **kwargs) -> CoAtNet:
    num_blocks = [2, 2, 6, 14, 2]
    channels = [128, 128, 256, 512, 1024]
    return CoAtNet((224, 224), 3, num_blocks, channels, num_classes, **kwargs)


@register_model
def coatnet_3(num_classes: int = 1000, **kwargs) -> CoAtNet:
    num_blocks = [2, 2, 6, 14, 2]
    channels = [192, 192, 384, 768, 1536]
    return CoAtNet((224, 224), 3, num_blocks, channels, num_classes, **kwargs)


@register_model
def coatnet_4(num_classes: int = 1000, **kwargs) -> CoAtNet:
    num_blocks = [2, 2, 12, 28, 2]
    channels = [192, 192, 384, 768, 1536]
    return CoAtNet((224, 224), 3, num_blocks, channels, num_classes, **kwargs)


@register_model
def coatnet_5(num_classes: int = 1000, **kwargs) -> CoAtNet:
    num_blocks = [2, 2, 12, 28, 2]
    channels = [192, 256, 512, 1280, 2048]
    return CoAtNet(
        (224, 224), 3, num_blocks, channels, num_classes, num_heads=64, **kwargs
    )


@register_model
def coatnet_6(num_classes: int = 1000, **kwargs) -> CoAtNet:
    num_blocks = [2, 2, 4, 8, 2]
    channels = [192, 192, 384, 768, 2048]
    return CoAtNet(
        (224, 224),
        3,
        num_blocks,
        channels,
        num_classes,
        num_heads=128,
        _scaled_num_blocks=[8, 42],
        _scaled_channels=[768, 1536],
        **kwargs
    )


@register_model
def coatnet_7(num_classes: int = 1000, **kwargs) -> CoAtNet:
    num_blocks = [2, 2, 4, 8, 2]
    channels = [192, 256, 512, 1024, 3072]
    return CoAtNet(
        (224, 224),
        3,
        num_blocks,
        channels,
        num_classes,
        num_heads=128,
        _scaled_num_blocks=[8, 42],
        _scaled_channels=[1024, 2048],
        **kwargs
    )



from typing import ClassVar

import lucid.nn as nn

from lucid import register_model
from lucid._tensor import Tensor

from .resnet import ResNet, _Bottleneck


__all__ = [
    "SENet",
    "se_resnet_18",
    "se_resnet_34",
    "se_resnet_50",
    "se_resnet_101",
    "se_resnet_152",
    "se_resnext_50_32x4d",
    "se_resnext_101_32x4d",
    "se_resnext_101_32x8d",
    "se_resnext_101_64x4d",
]


class SENet(ResNet):
    def __init__(
        self,
        block: nn.Module,
        layers: list[int],
        num_classes: int = 1000,
        reduction: int = 16,
        block_args: dict = {},
    ) -> None:
        super().__init__(
            block,
            layers,
            num_classes,
            block_args={"se": True, "se_args": dict(reduction=reduction), **block_args},
        )


class _SEResNetModule(nn.Module):
    expansion: ClassVar[int] = 1

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        stride: int = 1,
        downsample: nn.Module | None = None,
        reduction: int = 16,
        **kwargs,
    ) -> None:
        super().__init__()

        self.conv1 = nn.ConvBNReLU2d(
            in_channels,
            out_channels,
            kernel_size=3,
            stride=stride,
            padding=1,
            conv_bias=False,
        )
        self.conv2 = nn.Sequential(
            nn.Conv2d(
                out_channels,
                out_channels,
                kernel_size=3,
                padding=1,
                bias=False,
            ),
            nn.BatchNorm2d(out_channels),
        )

        self.se_module = nn.SEModule(out_channels, reduction)
        self.relu = nn.ReLU()
        self.downsample = downsample

    def forward(self, x: Tensor) -> Tensor:
        out = self.conv1(x)
        out = self.conv2(out)

        out = self.se_module(out)
        if self.downsample is not None:
            out += self.downsample(x)

        out = self.relu(out)
        return out


@register_model
def se_resnet_18(num_classes: int = 1000, **kwargs) -> SENet:
    layers = [2, 2, 2, 2]
    return SENet(_SEResNetModule, layers, num_classes, **kwargs)


@register_model
def se_resnet_34(num_classes: int = 1000, **kwargs) -> SENet:
    layers = [3, 4, 6, 3]
    return SENet(_SEResNetModule, layers, num_classes, **kwargs)


@register_model
def se_resnet_50(num_classes: int = 1000, **kwargs) -> SENet:
    layers = [3, 4, 6, 3]
    return SENet(_Bottleneck, layers, num_classes, **kwargs)


@register_model
def se_resnet_101(num_classes: int = 1000, **kwargs) -> SENet:
    layers = [3, 4, 23, 3]
    return SENet(_Bottleneck, layers, num_classes, **kwargs)


@register_model
def se_resnet_152(num_classes: int = 1000, **kwargs) -> SENet:
    layers = [3, 8, 36, 3]
    return SENet(_Bottleneck, layers, num_classes, **kwargs)


@register_model
def se_resnext_50_32x4d(num_classes: int = 1000, **kwargs) -> SENet:
    layers = [3, 4, 6, 3]
    block_args = {"cardinality": 32, "base_width": 4}
    return SENet(_Bottleneck, layers, num_classes, block_args=block_args, **kwargs)


@register_model
def se_resnext_101_32x4d(num_classes: int = 1000, **kwargs) -> SENet:
    layers = [3, 4, 23, 3]
    block_args = {"cardinality": 32, "base_width": 4}
    return SENet(_Bottleneck, layers, num_classes, block_args=block_args, **kwargs)


@register_model
def se_resnext_101_32x8d(num_classes: int = 1000, **kwargs) -> SENet:
    layers = [3, 4, 23, 3]
    block_args = {"cardinality": 32, "base_width": 8}
    return SENet(_Bottleneck, layers, num_classes, block_args=block_args, **kwargs)


@register_model
def se_resnext_101_64x4d(num_classes: int = 1000, **kwargs) -> SENet:
    layers = [3, 4, 23, 3]
    block_args = {"cardinality": 64, "base_width": 4}
    return SENet(_Bottleneck, layers, num_classes, block_args=block_args, **kwargs)



import math
from typing import Callable, Literal

from lucid.optim import Optimizer
from lucid.optim.lr_scheduler import LRScheduler


__all__ = [
    "LambdaLR",
    "StepLR",
    "MultiStepLR",
    "ExponentialLR",
    "CosineAnnealingLR",
    "ReduceLROnPlateau",
    "CyclicLR",
]


class LambdaLR(LRScheduler):
    def __init__(
        self,
        lr_lambda: Callable[[int], float],
        optimizer: Optimizer,
        last_epoch: int = -1,
        verbose: bool = False,
    ) -> None:
        if not callable(lr_lambda):
            raise TypeError(
                "lr_lambda must be a callable function "
                + "that takes an epoch index and returns a scaling factor."
            )

        self.lr_lambda = lr_lambda
        super().__init__(optimizer, last_epoch, verbose)

    def get_lr(self) -> list[float]:
        factor = self.lr_lambda(self.last_epoch)
        return [base_lr * factor for base_lr in self.base_lrs]


class StepLR(LRScheduler):
    def __init__(
        self,
        optimizer: Optimizer,
        step_size: int,
        gamma: float = 0.1,
        last_epoch: int = -1,
        verbose: bool = False,
    ) -> None:
        if step_size <= 0:
            raise ValueError("step_size must be a positive integer.")

        self.step_size = step_size
        self.gamma = gamma
        super().__init__(optimizer, last_epoch, verbose)

    def get_lr(self) -> list[float]:
        factor = self.gamma ** (self.last_epoch // self.step_size)
        return [base_lr * factor for base_lr in self.base_lrs]


class MultiStepLR(LRScheduler):
    def __init__(
        self,
        optimizer: Optimizer,
        milestones: list[int],
        gamma: float = 0.1,
        last_epoch: int = -1,
        verbose: bool = False,
    ) -> None:
        if not milestones:
            raise ValueError("milestones must contain at least one epoch index.")
        if sorted(milestones) != milestones:
            raise ValueError("milestones must be a sorted list of increasing integers.")

        self.milestones = milestones
        self.gamma = gamma
        super().__init__(optimizer, last_epoch, verbose)

    def get_lr(self) -> list[float]:
        factor = self.gamma ** sum(self.last_epoch >= m for m in self.milestones)
        return [base_lr * factor for base_lr in self.base_lrs]


class ExponentialLR(LRScheduler):
    def __init__(
        self,
        optimizer: Optimizer,
        gamma: float,
        last_epoch: int = -1,
        verbose: bool = False,
    ) -> None:
        if gamma <= 0:
            raise ValueError("gamma must be a positive float.")

        self.gamma = gamma
        super().__init__(optimizer, last_epoch, verbose)

    def get_lr(self) -> list[float]:
        factor = self.gamma**self.last_epoch
        return [base_lr * factor for base_lr in self.base_lrs]


class CosineAnnealingLR(LRScheduler):
    def __init__(
        self,
        optimizer: Optimizer,
        T_max: int,
        eta_min: float = 0.0,
        last_epoch: int = -1,
        verbose: bool = False,
    ) -> None:
        if T_max <= 0:
            raise ValueError("T_max must be a positive integer.")

        self.T_max = T_max
        self.eta_min = eta_min
        super().__init__(optimizer, last_epoch, verbose)

    def get_lr(self) -> list[float]:
        return [
            self.eta_min
            + (base_lr - self.eta_min)
            * (1 + math.cos(math.pi * self.last_epoch / self.T_max))
            / 2
            for base_lr in self.base_lrs
        ]


class ReduceLROnPlateau(LRScheduler):
    def __init__(
        self,
        optimizer: Optimizer,
        mode: Literal["min", "max"],
        factor: float = 0.1,
        patience: int = 10,
        threshold: float = 1e-4,
        threshold_mode: Literal["rel", "abs"] = "rel",
        cooldown: int = 0,
        min_lr: float = 0.0,
        eps: float = 1e-8,
        verbose: bool = False,
    ) -> None:
        if factor >= 1.0:
            raise ValueError("factor should be < 1.0.")
        if mode not in {"min", "max"}:
            raise ValueError("mode should be in 'min' or 'max'.")
        if threshold_mode not in {"rel", "abs"}:
            raise ValueError("threshold_mode must be 'rel' or 'abs'.")

        self.mode = mode
        self.factor = factor
        self.patience = patience
        self.threshold = threshold
        self.threshold_mode = threshold_mode
        self.cooldown = cooldown
        self.min_lr = min_lr
        self.eps = eps
        self.verbose = verbose

        self.best = float("inf") if mode == "min" else float("-inf")
        self.num_bad_epochs = 0
        self.cooldown_counter = 0
        super().__init__(optimizer, last_epoch=-1, verbose=verbose)

    def get_lr(self) -> list[float]:
        return self._last_lr

    def step(self, metrics: float, epoch: int | None = None) -> None:
        if epoch is not None:
            self.last_epoch = epoch

        if self.is_better(metrics):
            self.best = metrics
            self.num_bad_epochs = 0
        else:
            self.num_bad_epochs += 1

        if self.num_bad_epochs > self.patience:
            self._reduce_lr()
            self.num_bad_epochs = 0
            self.cooldown_counter = self.cooldown

        if self.cooldown_counter > 0:
            self.cooldown_counter -= 1

    def is_better(self, metrics: float) -> bool:
        if self.threshold_mode == "rel":
            threshold_val = (
                self.best * (1 - self.threshold)
                if self.mode == "min"
                else self.best * (1 + self.threshold)
            )
        else:
            threshold_val = (
                self.best - self.threshold
                if self.mode == "min"
                else self.best + self.threshold
            )
        return (
            metrics < threshold_val if self.mode == "min" else metrics > threshold_val
        )

    def _reduce_lr(self) -> None:
        new_lrs = []
        for param_group in self.optimizer.param_groups:
            new_lr = max(param_group["lr"] * self.factor, self.min_lr)
            if param_group["lr"] - new_lr > self.eps:
                param_group["lr"] = new_lr

            new_lrs.append(new_lr)

        self._last_lr = new_lrs
        if self.verbose:
            print(f"Reducing learning rate to {new_lrs}.")


class CyclicLR(LRScheduler):
    def __init__(
        self,
        optimizer: Optimizer,
        base_lr: float,
        max_lr: float,
        step_size_up: int,
        step_size_down: int | None = None,
        mode: Literal["triangular", "triangular2", "exp_range"] = "triangular",
        gamma: float = 1.0,
        scale_fn: Callable[[int], float] | None = None,
        cycle_momentum: bool = True,
        last_epoch: int = -1,
        verbose: bool = False,
    ) -> None:
        if base_lr >= max_lr:
            raise ValueError("base_lr must be less than max_lr.")
        if mode not in {"triangular", "triangular2", "exp_range"}:
            raise ValueError("Invalid mode.")

        self.base_lr = base_lr
        self.max_lr = max_lr
        self.step_size_up = step_size_up
        self.step_size_down = step_size_down or step_size_up
        self.mode = mode
        self.gamma = gamma
        self.scale_fn = scale_fn
        self.cycle_momentum = cycle_momentum

        self.total_size = self.step_size_up + self.step_size_down
        super().__init__(optimizer, last_epoch, verbose)

    def get_lr(self) -> list[float]:
        cycle = self.last_epoch // self.total_size
        x = abs(self.last_epoch / self.step_size_up - 2 * cycle - 1)

        if self.mode == "triangular":
            scale_factor = 1.0
        elif self.mode == "triangular2":
            scale_factor = 1 / (2**cycle)
        elif self.mode == "exp_range":
            scale_factor = self.gamma**self.last_epoch
        else:
            scale_factor = self.scale_fn(cycle) if self.scale_fn else 1.0

        new_lrs = [
            self.base_lr + (self.max_lr - self.base_lr) * max(0, 1 - x) * scale_factor
            for _ in self.base_lrs
        ]
        return new_lrs



from abc import ABC, abstractmethod
from typing import Callable, Tuple, ClassVar
import functools

import lucid
import lucid.types as types
from lucid.types import _DeviceType, _NumPyArray, _MLXArray

from lucid._tensor import Tensor
from lucid._backend.metal import is_gpu_op


_GradFuncType = Callable[[None], Tuple[_NumPyArray | _MLXArray, ...]]

_ReturnGradFuncPair = Tuple[Tensor, _GradFuncType]

_FuncOpReturnType = Tuple[_ReturnGradFuncPair, ...]


def func_op(
    n_in: int | None,
    n_ret: int | None,
    has_gradient: bool = True,
    device: _DeviceType = "cpu",
) -> Callable:
    def decorator(func: Callable[..., _FuncOpReturnType]) -> Callable:
        @functools.wraps(func)
        def wrapper(op_self: operation, *args, **kwargs) -> Tuple[Tensor, ...]:
            tensors: Tuple[Tensor, ...] = tuple()
            requires_grad = False
            is_free = True

            if n_in is None:
                tensor_args = args
            else:
                if len(args) < n_in:
                    raise ValueError(
                        f"Expected at least {n_in} tensor arguments, got {len(args)}"
                    )
                tensor_args = args[:n_in]

            for arg in tensor_args:
                tensor = lucid._check_is_tensor(arg, device=device)
                tensors += (tensor,)
                requires_grad = requires_grad or tensor.requires_grad

                if tensor.is_free:
                    tensor.to(device)
                else:
                    is_free = False
                    if tensor.device != device:
                        raise RuntimeError(
                            f"{tensor.device} tensor of shape {tensor.shape} "
                            + f"passed for {device} operation"
                            + f"('{type(op_self).__name__}')."
                        )

            non_tensor_args = args[n_in:] if n_in is not None else ()
            new_args = (*tensors, *non_tensor_args)
            func_return_pairs = func(op_self, *new_args, **kwargs)

            if lucid.flops_enabled():
                op_self.flops = op_self.__flops__(*new_args, **kwargs)

            if n_ret is None:
                if not isinstance(func_return_pairs, tuple):
                    raise ValueError(
                        f"{func.__name__} should return multiple '_ReturnGradFuncPair'."
                    )
                num_returns = len(func_return_pairs)
            else:
                num_returns = n_ret

            if num_returns == 1:
                func_return_pairs: _FuncOpReturnType = (func_return_pairs,)

            results: Tuple[Tensor, ...] = tuple()
            for result, compute_grad in func_return_pairs:
                result.requires_grad = requires_grad and has_gradient
                result._op = op_self
                result.to(device)
                if is_free:
                    result.free()

                results += (result,)
                if not lucid.grad_enabled() and not lucid.flops_enabled():
                    continue

                def _backward_op(*, _func: Callable = compute_grad) -> None:
                    grads = _func()
                    if n_in == 1 or not isinstance(grads, tuple):
                        grads = (grads,)

                    if len(grads) != len(tensors):
                        raise ValueError(
                            f"Expected {len(tensors)} gradients, got {len(grads)}."
                        )

                    for tensor, grad in zip(tensors, grads):
                        new_grad = lucid._match_grad_shape(
                            tensor.data, grad, device=device
                        )
                        lucid._set_tensor_grad(tensor, new_grad)

                if result.requires_grad or lucid.flops_enabled():
                    result._prev = list(tensors)
                    result._backward_op = (
                        _backward_op if result.requires_grad else lambda: None
                    )

            return results if num_returns > 1 else results[0]

        return wrapper

    return decorator


def unary_func_op(has_gradient: bool = True, device: _DeviceType = "cpu") -> Callable:
    return func_op(1, 1, has_gradient=has_gradient, device=device)


def binary_func_op(has_gradient: bool = True, device: _DeviceType = "cpu") -> Callable:
    return func_op(2, 1, has_gradient=has_gradient, device=device)


def poly_func_op(has_gradient: bool = True, device: _DeviceType = "cpu") -> Callable:
    return func_op(None, 1, has_gradient=has_gradient, device=device)


class operation(ABC):
    __fallback__: ClassVar[bool] = False

    def __init__(self) -> None:
        self.result: Tensor | tuple[Tensor, ...] | None = None
        self._flops: int | None = None

    @abstractmethod
    def cpu(self, *args, **kwargs) -> _FuncOpReturnType: ...

    @abstractmethod
    def gpu(self, *args, **kwargs) -> _FuncOpReturnType: ...

    def __grad__(self, *args, **kwargs) -> _GradFuncType: ...

    def __grad_cpu__(self, *args, **kwargs) -> _GradFuncType: ...

    def __grad_gpu__(self, *args, **kwargs) -> _GradFuncType: ...

    @property
    def flops(self) -> int:
        if self._flops is None:
            raise ValueError(f"flops counting for {self} has not been executed.")
        return self._flops

    @flops.setter
    def flops(self, val: int) -> None:
        self._flops = val

    def __flops__(self, *args, **kwargs) -> int:
        return 0

    def __call__(self, *args, **kwargs) -> Tensor | tuple[Tensor, ...]:
        if is_gpu_op(*args):
            return self.gpu(*args, **kwargs)
        return self.cpu(*args, **kwargs)


def fallback(cls: type[operation]) -> type[operation]:
    cls.__fallback__ = True
    return cls



from typing import ClassVar
from lucid import register_model

import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor
from lucid.models.objdet.util import nms, iou, DetectionDict


__all__ = ["YOLO_V3", "yolo_v3", "yolo_v3_tiny"]


def _convblock(
    cin: int, cout: int, k: int, s: int = 1, p: int | None = None
) -> list[nn.Module]:
    return [
        nn.Conv2d(
            cin,
            cout,
            kernel_size=k,
            stride=s,
            padding=p if p is not None else "same",
            bias=False,
        ),
        nn.BatchNorm2d(cout),
        nn.LeakyReLU(0.1),
    ]


class _ResidualBlock(nn.Module):
    def __init__(self, channels: int) -> None:
        super().__init__()
        half_channels = channels // 2
        self.block = nn.Sequential(
            *_convblock(channels, half_channels, 1),
            *_convblock(half_channels, channels, 3, p=1),
        )

    def forward(self, x: Tensor) -> Tensor:
        return x + self.block(x)


class _DarkNet_53(nn.Module):
    def __init__(self, num_classes: int = 1000) -> None:
        super().__init__()
        self.layer1 = nn.Sequential(*_convblock(3, 32, 3))
        self.layer2 = nn.Sequential(
            *_convblock(32, 64, 3, s=2, p=1), _ResidualBlock(64)
        )
        self.layer3 = nn.Sequential(
            *_convblock(64, 128, 3, s=2, p=1),
            *[_ResidualBlock(128) for _ in range(2)],
        )
        self.layer4 = nn.Sequential(
            *_convblock(128, 256, 3, s=2, p=1),
            *[_ResidualBlock(256) for _ in range(8)],
        )
        self.layer5 = nn.Sequential(
            *_convblock(256, 512, 3, s=2, p=1),
            *[_ResidualBlock(512) for _ in range(8)],
        )
        self.layer6 = nn.Sequential(
            *_convblock(512, 1024, 3, s=2, p=1),
            *[_ResidualBlock(1024) for _ in range(4)],
        )

        self.num_classes = num_classes
        self.gap = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(1024, self.num_classes)

    def forward(
        self, x: Tensor, classification: bool = False
    ) -> Tensor | tuple[Tensor, ...]:
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)

        r1 = self.layer4(x)
        r2 = self.layer5(r1)
        r3 = self.layer6(r2)

        if classification:
            out = self.gap(r3)
            return self.fc(out)

        return r1, r2, r3


class _DarkNet_53_Tiny(nn.Module):
    def __init__(self, num_classes: int = 1000) -> None:
        super().__init__()
        self.layer_prev = nn.Sequential(
            *_convblock(3, 16, 3, p=1),
            *_convblock(16, 32, 3, s=2, p=1),
            *_convblock(32, 64, 3, s=2, p=1),
        )
        self.layer4 = nn.Sequential(*_convblock(64, 128, 3, s=2, p=1))
        self.layer5 = nn.Sequential(*_convblock(128, 256, 3, s=2, p=1))
        self.layer6 = nn.Sequential(*_convblock(256, 512, 3, s=2, p=1))

        self.num_classes = num_classes
        self.gap = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(512, self.num_classes)

    def forward(
        self, x: Tensor, classification: bool = False
    ) -> Tensor | tuple[Tensor, ...]:
        x = self.layer_prev(x)

        r1 = self.layer4(x)
        r2 = self.layer5(r1)
        r3 = self.layer6(r2)

        if classification:
            out = self.gap(r3)
            return self.fc(out)

        return r1, r2, r3


_default_anchors = [
    (10, 13),
    (16, 30),
    (33, 23),
    (30, 61),
    (62, 45),
    (59, 119),
    (116, 90),
    (156, 198),
    (373, 326),
]


class YOLO_V3(nn.Module):
    default_anchors: ClassVar[list[tuple[int, int]]] = _default_anchors

    def __init__(
        self,
        num_classes: int,
        anchors: list[tuple[int, int]] | None = None,
        image_size: int = 416,
        darknet: nn.Module | None = None,
        darknet_out_channels_arr: list[int] | None = None,
    ) -> None:
        super().__init__()
        self.num_classes = num_classes
        self.image_size = image_size

        if anchors is None:
            anchors = YOLO_V3.default_anchors
        self.anchors = nn.Buffer(anchors, dtype=lucid.Float32)
        self.anchor_masks = [[6, 7, 8], [3, 4, 5], [0, 1, 2]]

        self.darknet = _DarkNet_53() if darknet is None else darknet
        if darknet is None:
            dark_out_chs = [256, 512, 1024]
        else:
            if darknet_out_channels_arr is None:
                raise ValueError(
                    "If providing a custom darknet, you must also specify "
                    "`darknet_out_channels_arr`."
                )
            dark_out_chs = darknet_out_channels_arr

        self.head1_pre = nn.Sequential(
            *_convblock(dark_out_chs[2], 512, 1),
            *_convblock(512, 1024, 3, p=1),
            *_convblock(1024, 512, 1),
            *_convblock(512, 1024, 3, p=1),
            *_convblock(1024, 512, 1),
        )
        self.head1_post = nn.Sequential(
            *_convblock(512, 1024, 3, p=1),
            nn.Conv2d(
                1024,
                len(self.anchor_masks[0]) * (5 + self.num_classes),
                kernel_size=1,
            ),
        )
        self.route1 = nn.Sequential(
            *_convblock(512, 256, 1), nn.Upsample(scale_factor=2, mode="nearest")
        )

        self.head2_pre = nn.Sequential(
            *_convblock(dark_out_chs[1] + 256, 256, 1),
            *_convblock(256, 512, 3, p=1),
            *_convblock(512, 256, 1),
            *_convblock(256, 512, 3, p=1),
            *_convblock(512, 256, 1),
        )
        self.head2_post = nn.Sequential(
            *_convblock(256, 512, 3, p=1),
            nn.Conv2d(
                512,
                len(self.anchor_masks[1]) * (5 + self.num_classes),
                kernel_size=1,
            ),
        )
        self.route2 = nn.Sequential(
            *_convblock(256, 128, 1), nn.Upsample(scale_factor=2, mode="nearest")
        )

        self.head3_pre = nn.Sequential(
            *_convblock(dark_out_chs[0] + 128, 128, 1),
            *_convblock(128, 256, 3, p=1),
            *_convblock(256, 128, 1),
            *_convblock(128, 256, 3, p=1),
            *_convblock(256, 128, 1),
        )
        self.head3_post = nn.Sequential(
            *_convblock(128, 256, 3, p=1),
            nn.Conv2d(
                256,
                len(self.anchor_masks[2]) * (5 + self.num_classes),
                kernel_size=1,
            ),
        )

    def forward(self, x: Tensor) -> tuple[Tensor]:
        r1, r2, r3 = self.darknet(x)

        x = self.head1_pre(r3)
        route = x
        out1 = self.head1_post(x)

        x = self.route1(route)
        x = lucid.concatenate([x, r2], axis=1)
        x = self.head2_pre(x)
        route = x
        out2 = self.head2_post(x)

        x = self.route2(route)
        x = lucid.concatenate([x, r1], axis=1)
        x = self.head3_pre(x)
        out3 = self.head3_post(x)

        return out1, out2, out3

    def _loss_per_scale(self, pred: Tensor, target: Tensor, anchors: Tensor) -> Tensor:
        N = pred.shape[0]
        C = self.num_classes
        B = anchors.shape[0]
        H, W = pred.shape[2:]

        pred = pred.reshape(N, B, 5 + C, H, W).transpose((0, 3, 4, 1, 2))
        target = target.reshape(N, H, W, B, 5 + C)

        stride = self.image_size / H
        ahw = anchors / stride

        obj_mask = target[..., 4:5]
        noobj_mask = 1.0 - obj_mask

        pred_xy_logits = pred[..., 0:2]
        pred_wh_logits = pred[..., 2:4]
        obj_logits = pred[..., 4:5]
        cls_logits = pred[..., 5:]

        tgt_xy_off = target[..., 0:2]
        tgt_wh_log = target[..., 2:4]
        tgt_cls = target[..., 5:]

        pred_xy = F.sigmoid(pred_xy_logits)
        pred_wh = lucid.exp(pred_wh_logits) * ahw.reshape(1, 1, 1, B, 2)
        tgt_wh = lucid.exp(tgt_wh_log) * ahw.reshape(1, 1, 1, B, 2)

        gy, gx = lucid.meshgrid(lucid.arange(H), lucid.arange(W), indexing="ij")
        grid = lucid.stack([gx, gy], axis=-1).reshape(1, H, W, 1, 2)

        pred_xy_px = (pred_xy + grid) * stride
        pred_wh_px = pred_wh * stride

        pred_x1y1 = pred_xy_px - pred_wh_px / 2
        pred_x1y2 = pred_xy_px + pred_wh_px / 2
        pred_boxes_px = lucid.concatenate([pred_x1y1, pred_x1y2], axis=-1)

        tgt_xy_px = (tgt_xy_off + grid) * stride
        tgt_wh_px = tgt_wh * stride

        tgt_x1y1 = tgt_xy_px - tgt_wh_px / 2
        tgt_x1y2 = tgt_xy_px + tgt_wh_px / 2
        tgt_boxes_px = lucid.concatenate([tgt_x1y1, tgt_x1y2], axis=-1)

        pred_boxes_flat = pred_boxes_px.reshape(N, -1, 4)
        tgt_boxes_flat = tgt_boxes_px.reshape(N, -1, 4)
        obj_mask_flat = obj_mask.reshape(N, -1, 1)

        max_iou = lucid.zeros((N, pred_boxes_flat.shape[1], 1), dtype=lucid.Float32)
        ignore_thresh = 0.5
        for n in range(N):
            if obj_mask_flat[n].sum() > 0:
                gt_n = tgt_boxes_flat[n][(obj_mask_flat[n][:, 0] > 0.5)]
                if gt_n.shape[0] > 0:
                    ious = iou(pred_boxes_flat[n], gt_n)
                    max_iou[n] = lucid.max(ious, axis=1, keepdims=True)

        max_iou = max_iou.reshape(N, H, W, B, 1)
        noobj_mask = noobj_mask * (max_iou < ignore_thresh).astype(lucid.Float32)

        box_scale = 2.0 - (tgt_wh_px[..., 0] * tgt_wh_px[..., 1]) / (self.image_size**2)

        w = box_scale * obj_mask[..., 0]
        w = lucid.clip(w, 0.0, None)
        sw = lucid.sqrt(w).reshape(N, H, W, B, 1)

        loss_xy = F.mse_loss(
            F.sigmoid(pred_xy_logits) * sw, tgt_xy_off * sw, reduction="sum"
        )
        loss_wh = F.mse_loss(pred_wh_logits * sw, tgt_wh_log * sw, reduction="sum")

        num_obj = lucid.maximum(obj_mask.sum(), 1.0)

        obj_prob = F.sigmoid(obj_logits)
        cls_prob = F.sigmoid(cls_logits)

        loss_obj = F.binary_cross_entropy(
            obj_prob, lucid.ones_like(obj_prob), weight=obj_mask, reduction="sum"
        )
        loss_noobj = F.binary_cross_entropy(
            obj_prob, lucid.zeros_like(obj_prob), weight=noobj_mask, reduction="sum"
        )
        loss_cls = F.binary_cross_entropy(
            cls_prob, tgt_cls, weight=obj_mask, reduction="sum"
        )

        total_loss = loss_xy + loss_wh + loss_obj + loss_noobj + loss_cls
        return total_loss / num_obj

    def get_loss(self, x: Tensor, target: tuple[Tensor]) -> Tensor:
        preds = self.forward(x)
        loss = 0.0
        for p, t, mask in zip(preds, target, self.anchor_masks):
            anchors = self.anchors[mask]
            loss += self._loss_per_scale(p, t, anchors)

        return loss / x.shape[0]

    @lucid.no_grad()
    def predict(
        self, x: Tensor, conf_thresh: float = 0.5, iou_thresh: float = 0.5
    ) -> list[list[DetectionDict]]:
        N = x.shape[0]
        C = self.num_classes
        preds = self.forward(x)

        results: list[list[DetectionDict]] = []
        for i in range(N):
            boxes_list = []
            scores_list = []
            for p, mask in zip(preds, self.anchor_masks):
                B = len(mask)
                H, W = p.shape[2:]
                stride = self.image_size / H
                anchor = self.anchors[mask] / stride

                pi = p[i].reshape(B, 5 + C, H, W).transpose((1, 2, 3, 0))

                pred_xy = F.sigmoid(pi[..., 0:2])
                pred_wh = lucid.exp(pi[..., 2:4]) * anchor.reshape(1, 1, B, 2)
                pred_obj = F.sigmoid(pi[..., 4:5])
                pred_cls = F.sigmoid(pi[..., 5:])

                grid_y, grid_x = lucid.meshgrid(
                    lucid.arange(H), lucid.arange(W), indexing="ij"
                )
                grid = lucid.stack([grid_x, grid_y], axis=-1).reshape(H, W, 1, 2)

                box_xy = (pred_xy + grid) * stride
                box_wh = pred_wh * stride

                box_xy1 = box_xy - box_wh / 2
                box_xy2 = box_xy + box_wh / 2

                box_xy1 = lucid.clip(box_xy1, 0.0, self.image_size - 1)
                box_xy2 = lucid.clip(box_xy2, 0.0, self.image_size - 1)

                boxes = lucid.concatenate([box_xy1, box_xy2], axis=-1).reshape(-1, 4)
                scores = (pred_obj * pred_cls).reshape(-1, C)

                boxes_list.append(boxes)
                scores_list.append(scores)

            boxes_img = lucid.concatenate(boxes_list, axis=0)
            scores_img = lucid.concatenate(scores_list, axis=0)

            image_preds: list[DetectionDict] = []
            for cl in range(C):
                cls_scores = scores_img[:, cl]
                mask = cls_scores > conf_thresh
                if not mask.any():
                    continue

                cls_boxes = boxes_img[mask]
                cls_scores = cls_scores[mask]

                keep = nms(cls_boxes, cls_scores, iou_thresh)
                for j in keep:
                    image_preds.append(
                        {
                            "box": cls_boxes[j].tolist(),
                            "score": cls_scores[j].item(),
                            "class_id": cl,
                        }
                    )

            results.append(image_preds)

        return results


@register_model
def yolo_v3(num_classes: int = 80, **kwargs) -> YOLO_V3:
    return YOLO_V3(num_classes=num_classes, image_size=416, **kwargs)


@register_model
def yolo_v3_tiny(num_classes: int = 80, **kwargs) -> YOLO_V3:
    return YOLO_V3(
        num_classes=num_classes,
        image_size=416,
        darknet=_DarkNet_53_Tiny(),
        darknet_out_channels_arr=[128, 256, 512],
        **kwargs,
    )



from typing import Any
from abc import ABC, abstractmethod

from lucid.optim import Optimizer


class LRScheduler(ABC):
    def __init__(
        self, optimizer: Optimizer, last_epoch: int = -1, verbose: bool = False
    ) -> None:
        super().__init__()
        if not hasattr(optimizer, "param_groups"):
            raise TypeError(f"{type(optimizer).__name__} is not a valid optimizer.")

        self.optimizer = optimizer
        self.last_epoch = last_epoch
        self.verbose = verbose
        self.base_lrs = [group["lr"] for group in optimizer.param_groups]

        self._step_count = 0
        self._last_lr = [group["lr"] for group in optimizer.param_groups]

    @abstractmethod
    def get_lr(self) -> list[float]:
        raise NotImplementedError("get_lr must be implemented in subclasses.")

    def step(self, epoch: int | None = None) -> None:
        if epoch is not None:
            self.last_epoch = epoch
        else:
            self._step_count += 1
            self.last_epoch = self._step_count

        new_lrs = self.get_lr()
        for param_group, lr in zip(self.optimizer.param_groups, new_lrs):
            param_group["lr"] = lr

        self._last_lr = new_lrs

        if self.verbose:
            print(f"Epoch {self.last_epoch}: setting learning rates to {new_lrs}.")

    def state_dict(self) -> dict[str, Any]:
        return {
            "last_epoch": self.last_epoch,
            "base_lrs": self.base_lrs,
            "_step_count": self._step_count,
            "_last_lr": self._last_lr,
        }

    def load_state_dict(self, state_dict: dict[str, Any]) -> None:
        self.last_epoch = state_dict["last_epoch"]
        self.base_lrs = state_dict["base_lrs"]

        self._step_count = state_dict["_step_count"]
        self._last_lr = state_dict["_last_lr"]

        for param_group, lr in zip(self.optimizer.param_groups, self._last_lr):
            param_group["lr"] = lr

    @property
    def last_lr(self) -> list[float]:
        return self._last_lr



from typing import Any, ClassVar
import lucid.nn as nn

from lucid import register_model
from lucid._tensor import Tensor

from .resnet import ResNet


__all__ = [
    "SKNet",
    "sk_resnet_18",
    "sk_resnet_34",
    "sk_resnet_50",
    "sk_resnext_50_32x4d",
]


class SKNet(ResNet):
    def __init__(
        self,
        block: nn.Module,
        layers: list[int],
        num_classes: int = 1000,
        kernel_sizes: list[int] = [3, 5],
        base_width: int = 64,
        cardinality: int = 1,
        **resnet_args: Any,
    ) -> None:
        super().__init__(
            block,
            layers,
            num_classes,
            block_args={
                "kernel_sizes": kernel_sizes,
                "base_width": base_width,
                "cardinality": cardinality,
            },
            **resnet_args,
        )


class _SKResNetModule(nn.Module):
    expansion: ClassVar[int] = 1

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        stride: int = 1,
        downsample: nn.Module | None = None,
        kernel_sizes: list[int] = [3, 5],
        cardinality: int = 1,
        base_width: int = 64,
    ) -> None:
        super().__init__()
        width = int(out_channels * (base_width / 64.0)) * cardinality

        self.conv1 = nn.ConvBNReLU2d(
            in_channels, width, kernel_size=1, stride=1, conv_bias=False
        )
        self.sk_module = nn.SelectiveKernel(
            width, width, kernel_sizes=kernel_sizes, stride=stride, groups=cardinality
        )
        self.bn2 = nn.BatchNorm2d(width)
        self.conv3 = nn.Sequential(
            nn.Conv2d(
                width,
                out_channels * self.expansion,
                kernel_size=1,
                stride=1,
                bias=False,
            ),
            nn.BatchNorm2d(out_channels * self.expansion),
        )

        self.relu = nn.ReLU()
        self.downsample = downsample

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.relu(self.bn2(self.sk_module(out)))
        out = self.conv3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class _SKResNetBottleneck(nn.Module):
    expansion: ClassVar[int] = 4

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        stride: int = 1,
        downsample: nn.Module | None = None,
        kernel_sizes: list[int] = [3, 5],
        cardinality: int = 1,
        base_width: int = 64,
    ) -> None:
        super().__init__()
        width = int(out_channels * (base_width / 64.0)) * cardinality

        self.conv1 = nn.ConvBNReLU2d(
            in_channels, width, kernel_size=1, stride=1, conv_bias=False
        )

        self.sk_module = nn.SelectiveKernel(
            width, width, kernel_sizes=kernel_sizes, stride=stride, groups=cardinality
        )
        self.bn2 = nn.BatchNorm2d(width)

        self.conv3 = nn.Sequential(
            nn.Conv2d(
                width,
                out_channels * self.expansion,
                kernel_size=1,
                stride=1,
                bias=False,
            ),
            nn.BatchNorm2d(out_channels * self.expansion),
        )

        self.relu = nn.ReLU()
        self.downsample = downsample

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.relu(self.bn2(self.sk_module(out)))
        out = self.conv3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


@register_model
def sk_resnet_18(num_classes: int = 1000, **kwargs) -> SKNet:
    layers = [2, 2, 2, 2]
    return SKNet(_SKResNetModule, layers, num_classes, **kwargs)


@register_model
def sk_resnet_34(num_classes: int = 1000, **kwargs) -> SKNet:
    layers = [3, 4, 6, 3]
    return SKNet(_SKResNetModule, layers, num_classes, **kwargs)


@register_model
def sk_resnet_50(num_classes: int = 1000, **kwargs) -> SKNet:
    layers = [3, 4, 6, 3]
    return SKNet(_SKResNetBottleneck, layers, num_classes, **kwargs)


@register_model
def sk_resnext_50_32x4d(num_classes: int = 1000, **kwargs) -> SKNet:
    layers = [3, 4, 6, 3]
    return SKNet(
        _SKResNetBottleneck,
        layers,
        num_classes,
        cardinality=32,
        base_width=4,
        **kwargs,
    )



import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor


__all__ = ["DDPM"]


class _ResBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        time_emb_dim: int,
        dropout: float = 0.1,
    ) -> None:
        super().__init__()
        self.norm1 = nn.GroupNorm(num_groups=32, num_channels=in_channels)
        self.activation = nn.Swish()
        self.conv1 = nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1)

        self.time_proj = nn.Linear(time_emb_dim, out_channels)
        self.norm2 = nn.GroupNorm(num_groups=32, num_channels=out_channels)
        self.conv2 = nn.Conv2d(out_channels, out_channels, kernel_size=3, padding=1)
        self.dropout = nn.Dropout2d(dropout)

        self.residual = (
            nn.Conv2d(in_channels, out_channels, kernel_size=1)
            if in_channels != out_channels
            else nn.Identity()
        )

    def forward(self, x: Tensor, t_emb: Tensor) -> Tensor:
        h = self.conv1(self.activation(self.norm1(x)))
        h += self.time_proj(t_emb)[:, :, None, None]
        h = self.conv2(self.activation(self.norm2(h)))
        h = self.dropout(h)

        return h + self.residual(x)


class _AttentionBlock(nn.Module):
    def __init__(self, channels: int) -> None:
        super().__init__()
        self.norm = nn.GroupNorm(num_groups=32, num_channels=channels)
        self.qkv = nn.Conv2d(channels, channels * 3, kernel_size=1)
        self.proj = nn.Conv2d(channels, channels, kernel_size=1)

    def forward(self, x: Tensor) -> Tensor:
        B, C, H, W = x.shape
        h = self.norm(x)
        q, k, v = self.qkv(h).chunk(3, axis=1)

        q = q.reshape(B, C, H * W)
        k = k.reshape(B, C, H * W)
        v = v.reshape(B, C, H * W)

        attn = (q.mT @ k) / C**0.5
        attn = F.softmax(attn, axis=-1)

        out = v @ attn.mT
        out = out.reshape(B, C, H, W)

        return x + self.proj(out)


class _UNet(nn.Module):
    def __init__(
        self,
        in_channels: int = 3,
        out_channels: int = 3,
        base_channels: int = 128,
        channel_mults: tuple[int, ...] = (1, 2, 2),
        num_res_blocks: int = 2,
        attention_res: tuple[int, ...] = (16,),
        image_size: int = 32,
        time_emb_dim: int = 512,
        dropout: float = 0.1,
        use_sigmoid: bool = True,
    ) -> None:
        super().__init__()
        self.base_channels = base_channels
        self.channel_mults = channel_mults
        self.num_res_blocks = num_res_blocks
        self.attn_resolutions = set(attention_res)
        self.image_size = image_size
        self.time_emb_dim = time_emb_dim
        self.use_sigmoid = use_sigmoid
        self.dropout = dropout

        self.time_mlp = nn.Sequential(
            nn.Linear(time_emb_dim, time_emb_dim * 4),
            nn.Swish(),
            nn.Linear(time_emb_dim * 4, time_emb_dim),
        )
        self.init_conv = nn.Conv2d(in_channels, base_channels, kernel_size=3, padding=1)

        self.down_resblocks = nn.ModuleList()
        self.downsample = nn.ModuleList()
        self.down_attn = nn.ModuleDict()

        curr_channels = base_channels
        curr_res = image_size
        for i, mult in enumerate(channel_mults):
            out_ch = base_channels * mult
            blocks = nn.ModuleList()

            for _ in range(num_res_blocks):
                blocks.append(_ResBlock(curr_channels, out_ch, time_emb_dim, dropout))
                curr_channels = out_ch

            self.down_resblocks.append(blocks)
            if curr_res in self.attn_resolutions:
                self.down_attn[str(i)] = _AttentionBlock(curr_channels)

            if i < len(channel_mults) - 1:
                self.downsample.append(
                    nn.Conv2d(
                        curr_channels, curr_channels, kernel_size=3, stride=2, padding=1
                    )
                )
                curr_res //= 2

        self.mid_block1 = _ResBlock(curr_channels, curr_channels, time_emb_dim, dropout)
        self.mid_attn = _AttentionBlock(curr_channels)
        self.mid_block2 = _ResBlock(curr_channels, curr_channels, time_emb_dim, dropout)

        self.upsample = nn.ModuleList()
        self.up_resblocks = nn.ModuleList()
        self.up_attn = nn.ModuleDict()

        rev_mults = list(reversed(channel_mults))
        curr_res = curr_res
        for i, mult in enumerate(rev_mults):
            if i > 0:
                self.upsample.append(
                    nn.Sequential(
                        nn.Upsample(scale_factor=2, mode="nearest"),
                        nn.Conv2d(
                            curr_channels, curr_channels, kernel_size=3, padding=1
                        ),
                    )
                )
                curr_res *= 2

            out_ch = base_channels * mult
            blocks = nn.ModuleList()
            for j in range(num_res_blocks):
                in_ch = curr_channels + (base_channels * mult if j == 0 else 0)
                blocks.append(
                    _ResBlock(in_ch or curr_channels, out_ch, time_emb_dim, dropout)
                )
                curr_channels = out_ch

            self.up_resblocks.append(blocks)
            if curr_res in self.attn_resolutions:
                self.up_attn[str(i)] = _AttentionBlock(curr_channels)

        self.final_norm = nn.GroupNorm(num_groups=32, num_channels=curr_channels)
        self.final_act = nn.Swish()
        self.final_conv = nn.Conv2d(
            curr_channels, out_channels, kernel_size=3, padding=1
        )
        self.final_sigmoid = nn.Sigmoid()

    def time_embedding(self, t: Tensor) -> Tensor:
        half_dim = self.time_emb_dim // 2
        scale = lucid.log(10000.0) / (half_dim - 1)

        emb = lucid.arange(half_dim, device=t.device) * -scale
        emb = lucid.exp(emb)[None, :] * t[:, None]
        emb = lucid.concatenate([lucid.sin(emb), lucid.cos(emb)], axis=-1)

        return self.time_mlp(emb)

    def forward(self, x: Tensor, t: Tensor) -> Tensor:
        t_emb = self.time_embedding(t)
        x = self.init_conv(x)

        skips: list[Tensor] = []
        for i, blocks in enumerate(self.down_resblocks):
            for layer in blocks:
                x = layer(x, t_emb)
            if str(i) in self.down_attn:
                x = self.down_attn[str(i)](x)
            skips.append(x)
            if i < len(self.downsample):
                x = self.downsample[i](x)

        x = self.mid_block1(x, t_emb)
        x = self.mid_attn(x)
        x = self.mid_block2(x, t_emb)

        for i, blocks in enumerate(self.up_resblocks):
            if i > 0:
                x = self.upsample[i - 1](x)
            skip = skips.pop()

            x = lucid.concatenate([x, skip], axis=1)
            for layer in blocks:
                x = layer(x, t_emb)

            if str(i) in self.up_attn:
                x = self.up_attn[str(i)](x)

        x = self.final_norm(x)
        x = self.final_act(x)
        x = self.final_conv(x)

        if self.use_sigmoid:
            x = self.final_sigmoid(x)

        return x


class _GaussianDiffuser(nn.Module):
    def __init__(
        self, timesteps: int = 1000, beta_start: float = 1e-4, beta_end: float = 0.02
    ) -> None:
        super().__init__()
        self.timesteps = timesteps
        self.register_buffer("betas", lucid.linspace(beta_start, beta_end, timesteps))
        self.register_buffer("alphas", 1.0 - self.betas)

        self.register_buffer("alphas_cumprod", lucid.cumprod(self.alphas, axis=0))
        self.register_buffer(
            "alphas_cumprod_prev",
            lucid.concatenate([Tensor([1.0]), self.alphas_cumprod[:-1]], axis=0),
        )
        self.register_buffer(
            "posterior_var",
            self.betas * (1.0 - self.alphas_cumprod_prev) / (1.0 - self.alphas_cumprod),
        )

    def sample_timesteps(self, batch_size: int) -> Tensor:
        return lucid.random.randint(0, self.timesteps, (batch_size,))

    def add_noise(self, x_start: Tensor, t: Tensor, noise: Tensor) -> Tensor:
        shape = (t.shape[0],) + (1,) * (x_start.ndim - 1)
        sqrt_alpha = lucid.sqrt(self.alphas_cumprod)[t].reshape(*shape)
        sqrt_one_minus_alpha = lucid.sqrt(1.0 - self.alphas_cumprod)[t].reshape(*shape)

        return sqrt_alpha * x_start + sqrt_one_minus_alpha * noise

    def denoise(
        self, model: nn.Module, x: Tensor, t: Tensor, clip_denoised: bool
    ) -> Tensor:
        noise_pred = model(x, t)
        shape = (t.shape[0],) + (1,) * (x.ndim - 1)

        alpha_t = self.alphas[t].reshape(*shape)
        beta_t = self.betas[t].reshape(*shape)
        alpha_bar = self.alphas_cumprod[t].reshape(*shape)

        pred_x0 = (x - lucid.sqrt(1 - alpha_bar) * noise_pred) / lucid.sqrt(alpha_bar)
        if clip_denoised:
            pred_x0 = pred_x0.clip(0.0, 1.0)

        posterior_mean = (beta_t * pred_x0 + (1 - beta_t) * x) / (1.0 - alpha_t)
        posterior_var = self.posterior_var[t].reshape(*shape)

        noise = lucid.random.randn(*x.shape)
        nonzero_mask = (t != 0).astype(x.dtype).reshape(*shape)

        return posterior_mean + nonzero_mask * lucid.sqrt(posterior_var) * noise


class DDPM(nn.Module):
    def __init__(
        self,
        model: nn.Module | None = None,
        image_size: int = 32,
        channels: int = 3,
        timesteps: int = 1000,
        diffuser: nn.Module | None = None,
        clip_denoised: bool = True,
    ) -> None:
        super().__init__()
        self.model = model or _UNet(
            in_channels=channels, out_channels=channels, image_size=image_size
        )
        self.diffuser = diffuser or _GaussianDiffuser(timesteps)

        self.image_size = image_size
        self.channels = channels
        self.timesteps = timesteps
        self.clip_denoised = clip_denoised

    def forward(self, x: Tensor) -> Tensor:
        return self.get_loss(x)

    def get_loss(self, x: Tensor) -> Tensor:
        B = x.shape[0]
        t = self.diffuser.sample_timesteps(B).to(x.device)
        noise = lucid.random.randn(*x.shape).to(x.device)

        x_t = self.diffuser.add_noise(x, t, noise)
        noise_pred = self.model(x_t, t)

        return F.mse_loss(noise_pred, noise)

    def sample(self, batch_size: int) -> Tensor:
        shape = (batch_size, self.channels, self.image_size, self.image_size)
        x = lucid.random.randn(*shape).to(self.device)

        for t in reversed(range(self.timesteps)):
            t_batch = lucid.full((shape[0],), t, dtype=lucid.Int32).to(self.device)
            x = self.diffuser.denoise(self.model, x, t_batch, self.clip_denoised)

        return x



from typing import Literal, ClassVar, Type

import lucid
import lucid.nn as nn
from lucid._tensor import Tensor


__all__ = [
    "ConvBNReLU1d",
    "ConvBNReLU2d",
    "ConvBNReLU3d",
    "DepthSeparableConv1d",
    "DepthSeparableConv2d",
    "DepthSeparableConv3d",
    "SEModule",
    "SelectiveKernel",
]


_PaddingStr = Literal["same", "valid"]

_Conv = [nn.Conv1d, nn.Conv2d, nn.Conv3d]
_BN = [nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d]


class _ConvBNReLU(nn.Module):
    D: ClassVar[int | None] = None

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int | tuple[int, ...],
        stride: int | tuple[int, ...] = 1,
        padding: _PaddingStr | int | tuple[int, ...] = 0,
        dilation: int | tuple[int, ...] = 1,
        groups: int = 1,
        conv_bias: bool = True,
        eps: float = 1e-5,
        momentum: float | None = 0.1,
        bn_affine: bool = True,
        track_running_stats: bool = True,
    ) -> None:
        super().__init__()
        if self.D is None:
            raise ValueError("Must specify 'D' value.")

        self.conv: nn.Module = _Conv[self.D - 1](
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            conv_bias,
        )
        self.bn: nn.Module = _BN[self.D - 1](
            out_channels, eps, momentum, bn_affine, track_running_stats
        )
        self.relu = nn.ReLU()

    def forward(self, input_: Tensor) -> Tensor:
        return self.relu(self.bn(self.conv(input_)))


class ConvBNReLU1d(_ConvBNReLU):
    D: ClassVar[int] = 1


class ConvBNReLU2d(_ConvBNReLU):
    D: ClassVar[int] = 2


class ConvBNReLU3d(_ConvBNReLU):
    D: ClassVar[int] = 3


class _DepthSeparableConv(nn.Module):
    D: ClassVar[int | None] = None

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int | tuple[int, ...],
        stride: int | tuple[int, ...] = 1,
        padding: _PaddingStr | int | tuple[int, ...] = 0,
        dilation: int | tuple[int, ...] = 1,
        bias: bool = True,
    ) -> None:
        super().__init__()
        if self.D is None:
            raise ValueError("Must specify 'D' value.")

        self.depthwise = _Conv[self.D - 1](
            in_channels,
            in_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups=in_channels,
            bias=bias,
        )
        self.pointwise = _Conv[self.D - 1](
            in_channels,
            out_channels,
            kernel_size=1,
            bias=bias,
        )
        self.reversed = reversed

    def forward(self, input_: Tensor) -> Tensor:
        return self.pointwise(self.depthwise(input_))


class DepthSeparableConv1d(_DepthSeparableConv):
    D: ClassVar[int] = 1


class DepthSeparableConv2d(_DepthSeparableConv):
    D: ClassVar[int] = 2


class DepthSeparableConv3d(_DepthSeparableConv):
    D: ClassVar[int] = 3


class SEModule(nn.Module):
    def __init__(self, in_channels: int, reduction: int = 16) -> None:
        super().__init__()
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))

        self.fc1 = nn.Linear(in_channels, in_channels // reduction)
        self.fc2 = nn.Linear(in_channels // reduction, in_channels)

        self.relu = nn.ReLU()
        self.sigmoid = nn.Sigmoid()

    def forward(self, x: Tensor) -> Tensor:
        if x.ndim != 4:
            raise ValueError("Only supports 4D-tensors.")

        spatial_axes = (-1, -2)
        y = self.avgpool(x).squeeze(axis=spatial_axes)
        y = self.relu(self.fc1(y))
        y = self.sigmoid(self.fc2(y))

        y = y.unsqueeze(axis=spatial_axes)
        out = x * y
        return out


class SelectiveKernel(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_sizes: list[int],
        stride: int = 1,
        padding: _PaddingStr | None = None,
        groups: int = 1,
        reduction: int = 16,
    ) -> None:
        super().__init__()

        branches = [
            nn.Conv2d(
                in_channels,
                out_channels,
                kernel_size=ks,
                stride=stride,
                padding=(ks // 2 if padding is None else padding),
                groups=groups,
                bias=False,
            )
            for ks in kernel_sizes
        ]
        self.branches = nn.ModuleList(branches)

        self.attention = nn.Sequential(
            nn.AdaptiveAvgPool2d((1, 1)),
            nn.Conv2d(
                out_channels, out_channels // reduction, kernel_size=1, bias=False
            ),
            nn.BatchNorm2d(out_channels // reduction),
            nn.ReLU(),
            nn.Conv2d(
                out_channels // reduction,
                len(kernel_sizes),
                kernel_size=1,
                bias=False,
            ),
        )

        self.softmax = nn.Softmax(axis=1)

    def forward(self, x: Tensor) -> Tensor:
        if x.ndim != 4:
            raise ValueError("Only supports 4D-tensors.")

        branch_outs = [branch(x) for branch in self.branches]
        branch_outs = lucid.stack(branch_outs, axis=1)

        att_scores = self.attention(branch_outs.sum(axis=1))
        att_weights = self.softmax(att_scores).unsqueeze(axis=2)

        out = (branch_outs * att_weights).sum(axis=1)
        return out



from functools import partial
from types import ModuleType
import math
import numpy as np

import lucid
from lucid._tensor import Tensor
from lucid.types import _Gradient, _ShapeLike

from lucid._backend.core import (
    operation,
    binary_func_op,
    _FuncOpReturnType,
    _GradFuncType,
)
from lucid._backend.metal import mx


def _broadcast_flops(a: Tensor, b: Tensor) -> int:
    out_shape = np.broadcast_shapes(a.shape, b.shape)
    return int(np.prod(out_shape))


class add(operation):
    def __init__(self) -> None:
        super().__init__()

    @binary_func_op()
    def cpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(a.data + b.data)
        return self.result, self.__grad__

    @binary_func_op(device="gpu")
    def gpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.add(a.data, b.data))
        return self.result, self.__grad__

    def __grad__(self) -> _GradFuncType:
        return self.result.grad, self.result.grad

    def __flops__(self, a: Tensor, b: Tensor) -> int:
        return _broadcast_flops(a, b)


class sub(operation):
    def __init__(self) -> None:
        super().__init__()

    @binary_func_op()
    def cpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(a.data - b.data)
        return self.result, self.__grad__

    @binary_func_op(device="gpu")
    def gpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.subtract(a.data, b.data))
        return self.result, self.__grad__

    def __grad__(self) -> _GradFuncType:
        return self.result.grad, -self.result.grad

    def __flops__(self, a: Tensor, b: Tensor) -> int:
        return _broadcast_flops(a, b)


class multiply(operation):
    def __init__(self) -> None:
        super().__init__()

    @binary_func_op()
    def cpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(a.data * b.data)
        return self.result, partial(self.__grad__, a=a, b=b)

    @binary_func_op(device="gpu")
    def gpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.multiply(a.data, b.data))
        return self.result, partial(self.__grad__, a=a, b=b)

    def __grad__(self, a: Tensor, b: Tensor) -> _GradFuncType:
        return b.data * self.result.grad, a.data * self.result.grad

    def __flops__(self, a: Tensor, b: Tensor) -> int:
        return _broadcast_flops(a, b)


class truediv(operation):
    def __init__(self) -> None:
        super().__init__()

    @binary_func_op()
    def cpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(a.data / b.data)
        return self.result, partial(self.__grad__, a=a, b=b)

    @binary_func_op(device="gpu")
    def gpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.divide(a.data, b.data))
        return self.result, partial(self.__grad__, a=a, b=b)

    def __grad__(self, a: Tensor, b: Tensor) -> _GradFuncType:
        return (
            (1 / b.data) * self.result.grad,
            (-a.data / (b.data**2)) * self.result.grad,
        )

    def __flops__(self, a: Tensor, b: Tensor) -> int:
        return _broadcast_flops(a, b)


class _equal(operation):
    def __init__(self) -> None:
        super().__init__()

    @binary_func_op(has_gradient=False)
    def cpu(self, a: Tensor, b: Tensor) -> Tensor:
        self.result = Tensor(a.data == b.data).astype(bool)
        return self.result, partial(self.__grad__, lib=np)

    @binary_func_op(has_gradient=False, device="gpu")
    def gpu(self, a: Tensor, b: Tensor) -> Tensor:
        self.result = Tensor(a.data == b.data).astype(bool)
        return self.result, partial(self.__grad__, lib=mx)

    def __grad__(self, lib_: ModuleType) -> _GradFuncType:
        return lib_.array(0.0), lib_.array(0.0)


class _not_equal(operation):
    def __init__(self) -> None:
        super().__init__()

    @binary_func_op(has_gradient=False)
    def cpu(self, a: Tensor, b: Tensor) -> Tensor:
        self.result = Tensor(a.data != b.data).astype(bool)
        return self.result, partial(self.__grad__, lib=np)

    @binary_func_op(has_gradient=False, device="gpu")
    def gpu(self, a: Tensor, b: Tensor) -> Tensor:
        self.result = Tensor(a.data != b.data).astype(bool)
        return self.result, partial(self.__grad__, lib=np)

    def __grad__(self, lib_: ModuleType) -> _GradFuncType:
        return lib_.array(0.0), lib_.array(0.0)


class _greater(operation):
    def __init__(self) -> None:
        super().__init__()

    @binary_func_op(has_gradient=False)
    def cpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(a.data > b.data).astype(bool)
        return self.result, partial(self.__grad__, lib=np)

    @binary_func_op(has_gradient=False, device="gpu")
    def gpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(a.data > b.data).astype(bool)
        return self.result, partial(self.__grad__, lib=np)

    def __grad__(self, lib_: ModuleType) -> _GradFuncType:
        return lib_.array(0.0), lib_.array(0.0)


class _greater_or_equal(operation):
    def __init__(self) -> None:
        super().__init__()

    @binary_func_op(has_gradient=False)
    def cpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(a.data >= b.data).astype(bool)
        return self.result, partial(self.__grad__, lib=np)

    @binary_func_op(has_gradient=False, device="gpu")
    def gpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(a.data >= b.data).astype(bool)
        return self.result, partial(self.__grad__, lib=np)

    def __grad__(self, lib_: ModuleType) -> _GradFuncType:
        return lib_.array(0.0), lib_.array(0.0)


class _less(operation):
    def __init__(self) -> None:
        super().__init__()

    @binary_func_op(has_gradient=False)
    def cpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(a.data < b.data).astype(bool)
        return self.result, partial(self.__grad__, lib=np)

    @binary_func_op(has_gradient=False, device="gpu")
    def gpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(a.data < b.data).astype(bool)
        return self.result, partial(self.__grad__, lib=np)

    def __grad__(self, lib_: ModuleType) -> _GradFuncType:
        return lib_.array(0.0), lib_.array(0.0)


class _less_or_equal(operation):
    def __init__(self) -> None:
        super().__init__()

    @binary_func_op(has_gradient=False)
    def cpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(a.data <= b.data).astype(bool)
        return self.result, partial(self.__grad__, lib=np)

    @binary_func_op(has_gradient=False, device="gpu")
    def gpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(a.data <= b.data).astype(bool)
        return self.result, partial(self.__grad__, lib=np)

    def __grad__(self, lib_: ModuleType) -> _GradFuncType:
        return lib_.array(0.0), lib_.array(0.0)


class minimum(operation):
    def __init__(self) -> None:
        super().__init__()

    @binary_func_op()
    def cpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.minimum(a.data, b.data))
        return self.result, partial(self.__grad__, a=a, b=b)

    @binary_func_op(device="gpu")
    def gpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.minimum(a.data, b.data))
        return self.result, partial(self.__grad__, a=a, b=b)

    def __grad__(self, a: Tensor, b: Tensor) -> _GradFuncType:
        a_grad = (a.data <= b.data).astype(a.data.dtype)
        b_grad = (a.data > b.data).astype(b.data.dtype)

        return a_grad * self.result.grad, b_grad * self.result.grad

    def __flops__(self, a: Tensor, b: Tensor) -> int:
        return _broadcast_flops(a, b)


class maximum(operation):
    def __init__(self) -> None:
        super().__init__()

    @binary_func_op()
    def cpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.maximum(a.data, b.data))
        return self.result, partial(self.__grad__, a=a, b=b)

    @binary_func_op(device="gpu")
    def gpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.maximum(a.data, b.data))
        return self.result, partial(self.__grad__, a=a, b=b)

    def __grad__(self, a: Tensor, b: Tensor) -> _GradFuncType:
        a_grad = (a.data >= b.data).astype(a.data.dtype)
        b_grad = (a.data < b.data).astype(b.data.dtype)

        return a_grad * self.result.grad, b_grad * self.result.grad

    def __flops__(self, a: Tensor, b: Tensor) -> int:
        return _broadcast_flops(a, b)


class power(operation):
    def __init__(self) -> None:
        super().__init__()

    @binary_func_op()
    def cpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.power(a.data, b.data))
        return self.result, partial(self.__grad__, a=a, b=b, lib_=np)

    @binary_func_op(device="gpu")
    def gpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.power(a.data, b.data))
        return self.result, partial(self.__grad__, a=a, b=b, lib_=mx)

    def __grad__(self, a: Tensor, b: Tensor, lib_: ModuleType) -> _GradFuncType:
        a_grad = b.data * lib_.power(a.data, b.data - 1)
        b_grad = lib_.power(a.data, b.data) * lib_.log(a.data)

        return a_grad * self.result.grad, b_grad * self.result.grad

    def __flops__(self, a: Tensor, b: Tensor) -> int:
        return _broadcast_flops(a, b)


class dot(operation):
    def __init__(self) -> None:
        super().__init__()

    @binary_func_op()
    def cpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.dot(a.data, b.data))
        return self.result, partial(self.__grad_cpu__, a=a, b=b)

    @binary_func_op(device="gpu")
    def gpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        if a.ndim != 1 or b.ndim != 1:
            raise ValueError(f"Only 1D dot product is supported for Metal backend.")

        self.result = Tensor(mx.sum(a.data * b.data))
        return self.result, partial(self.__grad_gpu__, a=a, b=b)

    def __grad_cpu__(self, a: Tensor, b: Tensor) -> _GradFuncType:
        return self.result.grad.dot(b.data.mT), a.data.mT.dot(self.result.grad)

    def __grad_gpu__(self, a: Tensor, b: Tensor) -> _GradFuncType:
        return b.data * self.result.grad, a.data * self.result.grad

    def __flops__(self, a: Tensor, b: Tensor) -> int:
        if a.ndim == 1 and b.ndim == 1:
            return 2 * a.shape[0] - 1

        a_batch_dims = a.shape[:-2] if a.ndim > 2 else ()
        b_batch_dims = b.shape[:-2] if b.ndim > 2 else ()

        m = a.shape[-2] if a.ndim >= 2 else 1
        k = a.shape[-1] if a.ndim >= 1 else 1
        k_b = b.shape[-2] if b.ndim >= 2 else 1
        n = b.shape[-1] if b.ndim >= 2 else 1

        if k != k_b:
            raise ValueError("Incompatible shapes for dot product.")

        batch_size = 1
        if a_batch_dims or b_batch_dims:
            max_dims = max(len(a_batch_dims), len(b_batch_dims))
            a_padded = (1,) * (max_dims - len(a_batch_dims)) + a_batch_dims
            b_padded = (1,) * (max_dims - len(b_batch_dims)) + b_batch_dims

            for a_dim, b_dim in zip(a_padded, b_padded):
                batch_size *= max(a_dim, b_dim)

        return batch_size * m * n * (2 * k - 1)


class inner(operation):
    def __init__(self) -> None:
        super().__init__()

    @binary_func_op()
    def cpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.inner(a.data, b.data))
        return self.result, partial(self.__grad__, a=a, b=b, lib=np)

    @binary_func_op(device="gpu")
    def gpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.inner(a.data, b.data))
        return self.result, partial(self.__grad__, a=a, b=b, lib=mx)

    def __grad__(self, a: Tensor, b: Tensor, lib_: ModuleType) -> _GradFuncType:
        return (
            lib_.tensordot(self.result.grad, b.data, axes=([-1], [-1])),
            lib_.tensordot(a.data, self.result.grad, axes=([-1], [-1])),
        )

    def __flops__(self, a: Tensor, b: Tensor) -> int:
        if a.ndim == 1 and b.ndim == 1:
            n = a.shape[0]
            return 2 * n - 1

        elif a.ndim >= 1 and b.ndim >= 1:
            n = a.shape[-1]
            if b.shape[-1] != n:
                raise ValueError("Last dimensions must match for inner product")

            out_shape = list(a.shape[:-1]) + list(b.shape[:-1])
            output_size = math.prod(out_shape)
            return output_size * (2 * n - 1)


class outer(operation):
    def __init__(self) -> None:
        super().__init__()

    @binary_func_op()
    def cpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.outer(a.data, b.data))
        return self.result, partial(self.__grad__, a=a, b=b, lib_=np)

    @binary_func_op(device="gpu")
    def gpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.outer(a.data, b.data))
        return self.result, partial(self.__grad__, a=a, b=b, lib_=mx)

    def __grad__(self, a: Tensor, b: Tensor, lib_: ModuleType) -> _GradFuncType:
        return (
            lib_.tensordot(self.result.grad, b.data, axes=([1], [0])),
            lib_.tensordot(self.result.grad, a.data, axes=([0], [0])),
        )

    def __flops__(self, a: Tensor, b: Tensor) -> int:
        return a.size * b.size


class matmul(operation):
    def __init__(self) -> None:
        super().__init__()

    @binary_func_op()
    def cpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        out = np.matmul(a.data, b.data)
        self.result = Tensor(out)
        return self.result, partial(self.__grad__, a=a, b=b, lib_=np)

    @binary_func_op(device="gpu")
    def gpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        out = mx.matmul(a.data, b.data)
        self.result = Tensor(out)
        return self.result, partial(self.__grad__, a=a, b=b, lib_=mx)

    def __grad__(self, a: Tensor, b: Tensor, lib_: ModuleType) -> _GradFuncType:
        grad = self.result.grad
        if grad.ndim == 0:
            grad = lib_.reshape(grad, (1, 1))

        grad_a = lib_.matmul(grad, lib_.swapaxes(b.data, -1, -2))
        grad_b = lib_.matmul(lib_.swapaxes(a.data, -1, -2), grad)

        grad_a = self._reduce_broadcast_shape(grad_a, a.shape, lib_)
        grad_b = self._reduce_broadcast_shape(grad_b, b.shape, lib_)

        return grad_a, grad_b

    def _reduce_broadcast_shape(
        self, grad: _Gradient, ref_shape: _ShapeLike, lib_: ModuleType
    ) -> _Gradient:
        while len(grad.shape) > len(ref_shape):
            grad = grad.sum(axis=0, keepdims=False)

        for i, (gdim, rdim) in enumerate(zip(grad.shape, ref_shape)):
            if gdim != rdim:
                grad = grad.sum(axis=i, keepdims=True)

        grad = lib_.reshape(grad, ref_shape)
        return grad

    def __flops__(self, a: Tensor, b: Tensor) -> int:
        a_shape, b_shape = a.shape, b.shape

        m = a_shape[-2] if len(a_shape) >= 2 else 1
        k = a_shape[-1] if len(a_shape) >= 1 else 1
        n = b_shape[-1] if len(b_shape) >= 1 else 1

        a_batch = (
            (1,) * (len(b_shape) - len(a_shape)) + a_shape[:-2]
            if len(a_shape) > 2
            else (1,) * (len(b_shape) - len(a_shape))
        )
        b_batch = (
            (1,) * (len(a_shape) - len(b_shape)) + b_shape[:-2]
            if len(b_shape) > 2
            else (1,) * (len(a_shape) - len(b_shape))
        )
        batch_shape = [max(x, y) for x, y in zip(a_batch, b_batch)]
        batch_size = np.prod(batch_shape) if batch_shape else 1

        return batch_size * m * n * k


class tensordot(operation):
    def __init__(
        self, axes: int | tuple[int, int] | tuple[list[int], list[int]]
    ) -> None:
        super().__init__()
        self.axes = axes

    def _get_axes_lists(self) -> tuple[list[int], list[int]]:
        if isinstance(self.axes, int):
            return [self.axes], [self.axes]

        if isinstance(self.axes, tuple):
            if len(self.axes) == 2 and all(isinstance(x, int) for x in self.axes):
                return [self.axes[0]], [self.axes[1]]
            elif len(self.axes) == 2 and all(isinstance(x, list) for x in self.axes):
                return list(self.axes[0]), list(self.axes[1])
            else:
                raise ValueError("Invalid axes format for tensordot operation.")

    @binary_func_op()
    def cpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        axes_a, axes_b = self._get_axes_lists()
        self.result = Tensor(np.tensordot(a.data, b.data, axes=(axes_a, axes_b)))
        return self.result, partial(self.__grad__, a=a, b=b, lib_=np)

    @binary_func_op(device="gpu")
    def gpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        axes_a, axes_b = self._get_axes_lists()
        self.result = Tensor(mx.tensordot(a.data, b.data, axes=(axes_a, axes_b)))
        return self.result, partial(self.__grad__, a=a, b=b, lib_=mx)

    def __grad__(self, a: Tensor, b: Tensor, lib_: ModuleType) -> _GradFuncType:
        axes_a, axes_b = self._get_axes_lists()
        grad = self.result.grad

        free_axes_a = [i for i in range(a.ndim) if i not in axes_a]
        free_axes_b = [i for i in range(b.ndim) if i not in axes_b]

        grad_a = lib_.tensordot(
            grad, b.data, axes=(list(range(len(free_axes_a), grad.ndim)), free_axes_b)
        )
        perm_a = free_axes_a + axes_a
        inv_perm_a = [perm_a.index(i) for i in range(a.ndim)]
        grad_a = lib_.transpose(grad_a, inv_perm_a)

        grad_b = lib_.tensordot(
            a.data, grad, axes=(free_axes_a, list(range(len(free_axes_a))))
        )
        perm_b = axes_b + free_axes_b
        inv_perm_b = [perm_b.index(i) for i in range(b.ndim)]
        grad_b = lib_.transpose(grad_b, inv_perm_b)

        return grad_a, grad_b

    def __flops__(self, a: Tensor, b: Tensor) -> int:
        axes_a, axes_b = self._get_axes_lists()
        m = math.prod(a.shape[i] for i in axes_a)
        n = math.prod(b.shape[i] for i in axes_b)

        num_out = (a.size // m) * (b.size // n)
        return num_out * (2 * m - 1)


class _bitwise_and(operation):
    def __init__(self) -> None:
        super().__init__()

    @binary_func_op(has_gradient=False)
    def cpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.bitwise_and(a.data, b.data))
        return self.result, partial(self.__grad__, lib_=np)

    @binary_func_op(has_gradient=False, device="gpu")
    def gpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.bitwise_and(a.data, b.data))
        return self.result, partial(self.__grad__, lib_=mx)

    def __grad__(self, lib_: ModuleType) -> _GradFuncType:
        return lib_.array(0.0), lib_.array(0.0)


class _bitwise_or(operation):
    def __init__(self) -> None:
        super().__init__()

    @binary_func_op(has_gradient=False)
    def cpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.bitwise_or(a.data, b.data))
        return self.result, partial(self.__grad__, lib_=np)

    @binary_func_op(has_gradient=False, device="gpu")
    def gpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.bitwise_or(a.data, b.data))
        return self.result, partial(self.__grad__, lib_=mx)

    def __grad__(self, lib_: ModuleType) -> _GradFuncType:
        return lib_.array(0.0), lib_.array(0.0)
